def is_alphanumeric(word, valid_punctuation_marks='-'):
    """
    Check if a word contains only alpha-numeric
    characters and valid punctuation marks.

    Parameters
    ----------
    word: string
        The given word

    valid_punctuation_marks: str
        Punctuation marks that are valid, defaults to '-'.

    Returns
    -------
    result: bool
        The result
    """

    for punctuation_mark in valid_punctuation_marks.split():
        word = word.replace(punctuation_mark, '')
    return word.isalnum()



class WordNormalizationMethod:
    """
    Represents a word normalization method.
    """
    stemming = 'stemming'
    lemmatization = 'lemmatization'


class TopicHeuristic:
    """
    Represents a heuristic for selecting the best candidate for each topic.
    """
    first_occurring = 'first_occurring'
    frequent = 'frequent'


class HierarchicalClusteringLinkageMethod:
    """
    Represents a hierarchical clustering linkage method
    """
    single = 'single'
    complete = 'complete'
    average = 'average'


class HierarchicalClusteringMetric:
    """
    Represents a hierarchical clustering metric
    """
    euclidean = 'euclidean'
    seuclidean = 'seuclidean'
    jaccard = 'jaccard'



class Sentence:
    """
    The sentence data structure

    Attributes
    ----------
    words: list
        List of words

    pos_tags: list
        List of part of speech tags assigned to words

    normalized_words: list
        List of normalized of words
    """

    def __init__(self, words, pos_tags, normalized_words):
        """
        Initializes the sentence.

        Parameters
        ----------
        words: list
            List of words

        pos_tags: list
            List of part of speech tags assigned to words

        normalized_words: list
            List of normalized of words
        """

        self.words = words
        self.pos_tags = pos_tags
        self.normalized_words = normalized_words

    @property
    def length(self):
        """
        Gets number of words

        Returns
        -------
        length: int
            Number of words
        """

        return len(self.words)


class Candidate:
    """
    The keyphrase candidate data structure

    Attributes
    ----------
    all_words: list
        Nested list of words, each words list in the list corresponds to
        one of the candidate occurrence.

    offsets: list
        List of offsets of each occurrence.

    all_pos_tags: list
        Nested list of pos tags, each pos tags list in the list
        corresponds to one of the candidate occurrence.

    normalized_words: list
        List of normalized of words, all occurrences have the same
        list of normalized words.

    weight: float
        Candidate weight in weighting algorithms.
    """

    def __init__(self):
        """
        Initializes the candidate.
        """

        self.all_words = []
        self.offsets = []
        self.all_pos_tags = []
        self.normalized_words = []
        self.weight = 0.

    @property
    def length(self):
        """
        Gets number of normalized words

        Returns
        -------
        length: int
            Number of normalized words
        """

        return len(self.normalized_words)

    def add_occurrence(self, words, offset, pos_tags, normalized_words):
        """
        Adds a new occurrence to the candidate.

        Parameters
        ----------
        words: list
            List of words of the occurrence

        normalized_words: list
            List of normalized of words of the occurrence

        pos_tags: list
            List of part of speech tags assigned to words of the
            occurrence

        offset: int
            The offset of the occurrence
        """
        self.all_words.append(words)
        self.offsets.append(offset)
        self.all_pos_tags.append(pos_tags)
        self.normalized_words = normalized_words



from collections import defaultdict

import networkx as nx

from perke.unsupervised.graph_based.single_rank import SingleRank


class PositionRank(SingleRank):
    """
    PositionRank keyphrase extractor

    This model is an unsupervised approach to extract keyphrases from 
    scholarly documents that incorporates information from all positions 
    of a word's occurrences into a biased PageRank.

    Implementation of the PositionRank described in:

    Corina Florescu and Cornelia Caragea.
    PositionRank: An Unsupervised Approach to Keyphrase Extraction from 
    Scholarly Documents.
    In proceedings of ACL, pages 1105-1115, 2017.

    Examples
    --------
    from perke.unsupervised.graph_based import PositionRank

    # Define the set of valid part of speech tags to occur in the model.
    valid_pos_tags = {'N', 'Ne', 'AJ', 'AJe'}

    # Define the grammar for selecting the keyphrase candidates
    grammar = r'
        NP:
            <P>{<N>}<V>
        NP:
            {<DETe?|Ne?|NUMe?|AJe|PRO|CL|RESe?><DETe?|Ne?|NUMe?|AJe?|PRO|CL|RESe?>*}
            <N>}{<.*e?>
    '

    # 1. Create a PositionRank extractor.
    extractor = PositionRank(valid_pos_tags=valid_pos_tags)

    # 2. Load the text.
    extractor.load_text(input='text or path/to/input_file', 
                        word_normalization_method=None)

    # 3. Select the noun phrases up to 3 words as keyphrase candidates.
    extractor.select_candidates(grammar=grammar, maximum_word_number=3)

    # 4. Weight the candidates using the sum of their word's weights
    #    that are computed using random walk biased with the position of
    #    the words in the text. In the graph, nodes are words (nouns
    #    and adjectives only) that are connected if they co-occur in a
    #    window of 10 words.
    extractor.weight_candidates(window_size=10)

    # 5. Get the 10 highest weighted candidates as keyphrases
    keyphrases = extractor.get_n_best(n=10)

    Attributes
    ----------
    positions: defaultdict(float)
        Container the sums of word's inverse positions
    """

    def __init__(self, valid_pos_tags=None):
        """
        Initializes PositionRank.

        Parameters
        ----------
        valid_pos_tags: set
            Set of valid part of speech tags, defaults to nouns and
            adjectives. I.e. `{'N', 'Ne', 'AJ', 'AJe'}`.
        """
        super().__init__(valid_pos_tags)
        self.positions = defaultdict(float)

    def select_candidates(self, grammar=None, maximum_length=3, **kwargs):
        """
        Candidate selection heuristic using a syntactic part of speech 
        pattern for noun phrase extraction. Keyphrase candidates are 
        noun phrases that match the regular expression 
        (adjective)*(noun)+, with a given length.

        Parameters
        ----------
        grammar: str
            Grammar defining part of speech patterns of noun phrases, 
            defaults to
            `'NP:
                <P>{<N>}<V>
            NP:
                {<DETe?|Ne?|NUMe?|AJe|PRO|CL|RESe?><DETe?|Ne?|NUMe?
                |AJe?|PRO|CL|RESe?>*}
                <N>}{<.*e?>.'`

        maximum_length: int
            Maximum length in words of the candidate, defaults to `3`.
        """

        if grammar is None:
            grammar = r"""
                NP:
                    <P>{<N>}<V>
                NP:
                    {<DETe?|Ne?|NUMe?|AJe|PRO|CL|RESe?><DETe?|Ne?|NUMe?|AJe?|PRO|CL|RESe?>*}
                    <N>}{<.*e?>
            """

        # Select sequence of noun phrases with given pattern
        self.select_candidates_with_grammar(grammar=grammar)

        self.filter_candidates(maximum_length)

    def build_word_graph(self, window_size=10):
        """
        Build the graph representation of the text. In the graph, nodes
        are words that passes a parts of speech filter. Two nodes are
        connected if the words corresponding to these nodes co-occur
        within a window of contiguous tokens. The weight of an edge is
        computed based on the co-occurrence count of the two words
        within a window of successive tokens.

        Parameters
        ----------
        window_size: int
            The size of window for connecting two words in the graph,
            defaults to `10`.
        """

        # Flatten text as a sequence of only passed syntactic filter
        # (word, position) tuples
        flatten_text = []
        shift = 0
        for i, sentence in enumerate(self.sentences):
            for j, word in enumerate(sentence.normalized_words):
                if sentence.pos_tags[j] in self.valid_pos_tags:
                    flatten_text.append((word, shift + j))

                    # Add node to the graph
                    self.graph.add_node(word)

            shift += sentence.length

        # Add edges to the graph
        for i, (first_node, first_node_position) in enumerate(flatten_text):
            for second_node, second_node_position in flatten_text[i + 1:]:
                if ((second_node_position - first_node_position) < window_size
                        and first_node != second_node):
                    if not self.graph.has_edge(first_node, second_node):
                        self.graph.add_edge(first_node, second_node, weight=0)

                    self.graph[first_node][second_node]['weight'] += 1

        # Compute the sums of the word's inverse positions
        for word, position in flatten_text:
            self.positions[word] += 1 / (position + 1)

    def weight_candidates(self,
                          window_size=10,
                          normalize_weights=False,
                          **kwargs):
        """
        Calculates candidates weights using a biased PageRank.

        Parameters
        ----------
        window_size: int
            The size of window for connecting two words in the graph,
            defaults to `10`.

        normalize_weights: bool
            Normalize keyphrase weight by their length, defaults to
            `False`.
        """

        # Build the word graph
        self.build_word_graph(window_size)

        # Normalize cumulated inverse positions
        position_sum = sum(self.positions.values())
        for word in self.positions:
            self.positions[word] /= position_sum

        # Compute the word weights using biased random walk
        weights = nx.pagerank(G=self.graph,
                              alpha=0.85,
                              tol=0.0001,
                              personalization=self.positions,
                              weight='weight')

        self.weight_candidates_with_words_weights(
            weights,
            normalize_weights,
            use_position_adjustment=False)

    def filter_candidates(self, maximum_length=3, **kwargs):
        """
        Filters the candidates with given conditions.

        Parameters
        ----------
        maximum_length: int
            Maximum length in words of the candidate, defaults to 3.
        """
        for c in list(self.candidates):
            if len(self.candidates[c].normalized_words) > maximum_length:
                del self.candidates[c]



from perke.unsupervised.graph_based.text_rank import TextRank
from perke.unsupervised.graph_based.single_rank import SingleRank
from perke.unsupervised.graph_based.position_rank import PositionRank
from perke.unsupervised.graph_based.topic_rank import TopicRank
from perke.unsupervised.graph_based.multipartite_rank import MultipartiteRank



import networkx as nx

from perke.base.extractor import Extractor


class TextRank(Extractor):
    """
    TextRank keyphrase extractor

    This model builds a graph that represents the text. A graph based
    ranking algorithm is then applied to extract the phrases that are
    most important in the text.

    In this implementation, nodes are words of certain parts of speech
    (nouns and adjectives) and edges represent co-occurrence relation,
    controlled by the distance between word occurrences
    (here a window of 2 words). Nodes are weighted by the TextRank
    graph-based weighting algorithm in its unweighted variant.

    Implementation of the TextRank model for keyword extraction
    described in:

    Rada Mihalcea and Paul Tarau.
    TextRank: Bringing Order into Texts
    In Proceedings of EMNLP, 2004.

    Examples
    --------
    from perke.unsupervised.graph_based import TextRank

    # Define the set of valid part of speech tags to occur in the model.
    valid_pos_tags = {'N', 'Ne', 'AJ', 'AJe'}

    # 1. Create a TextRank extractor.
    extractor = TextRank(valid_pos_tags=valid_pos_tags)

    # 2. Load the text.
    extractor.load_text(input='text or path/to/input_file',
                        word_normalization_method=None)

    # 3. Build the graph representation of the text and weight the
    #    words. Keyphrase candidates are composed from the 33 percent
    #    highest weighted words.
    extractor.weight_candidates(window_size=2, top_t_percent=0.33)

    # 4. Get the 10 highest weighted candidates as keyphrases.
    keyphrases = extractor.get_n_best(n=10)

    Attributes
    ----------
    graph: nx.Graph
        The word graph
    """

    def __init__(self, valid_pos_tags=None):
        """
        Initializes TextRank.

        Parameters
        ----------
        valid_pos_tags: set
            Set of valid part of speech tags, defaults to nouns and
            adjectives. I.e. `{'N', 'Ne', 'AJ', 'AJe'}`.
        """
        super().__init__(valid_pos_tags)
        self.graph = nx.Graph()
        self.graph_edges_are_weighted = False

    def select_candidates(self):
        """
        Selects candidates using longest sequences of certain parts of
        speech.
        """

        # Select sequences of words with valid pos tags
        self.select_candidates_with_longest_pos_sequences(
            valid_pos_tags=self.valid_pos_tags)

    def build_word_graph(self, window_size=2):
        """
        Builds a graph representation of the text in which nodes are
        words and edges represent co-occurrence relation. Syntactic
        filters can be applied to select only words with certain
        part of speech tags (noun and adjectives by default).
        Co-occurrence relations can be controlled using the distance
        between word occurrences (window) in the text.

        As the original paper does not give precise details on how the
        word graph is constructed, we make the following assumptions
        from the example given in Figure 2:
            1. Sentence boundaries **are not** taken into account.
            2. Stopwords and punctuation marks **are** considered as
               words when computing the window.

        Parameters
        ----------
        window_size: int
            The size of window for connecting two words in the graph,
            defaults to `2`.
        """

        # Flatten text as a sequence of (word, is_valid) tuples
        flatten_text = []
        for sentence in self.sentences:
            for i, word in enumerate(sentence.normalized_words):
                is_valid = sentence.pos_tags[i] in self.valid_pos_tags
                flatten_text.append((word, is_valid))

                # Add node to the graph
                if is_valid:
                    self.graph.add_node(word)

        # Add edges to the graph
        for i, (first_node, first_node_is_valid) in enumerate(flatten_text):

            # Speed up things
            if not first_node_is_valid:
                continue

            for j in range(i + 1, min(i + window_size, len(flatten_text))):
                second_node, second_node_is_valid = flatten_text[j]

                if second_node_is_valid and first_node != second_node:

                    # SingleRank
                    if self.graph_edges_are_weighted:
                        if not self.graph.has_edge(first_node, second_node):
                            self.graph.add_edge(first_node,
                                                second_node,
                                                weight=0.0)

                        self.graph[first_node][second_node]['weight'] += 1.0

                    # TextRank
                    else:
                        self.graph.add_edge(first_node, second_node)

    def weight_candidates(self, window_size=2,
                          top_t_percent=None,
                          normalize_weights=False):
        """
        Tailored candidate weighting method for TextRank. Keyphrase
        candidates are either composed from the top T highest weighted
        words as in the original paper or extracted using the
        `select_candidates` method. Candidates are weighting using the
        sum of their (normalized?) words.

        Parameters
        ----------
        window_size: int
            The size of window for connecting two words in the graph,
            defaults to `2`.

        top_t_percent: float
            Percentage of top vertices to keep for phrase generation.

        normalize_weights: bool
            Whether normalize keyphrase weight by their length, defaults
            to `False`.
        """

        # Build the word graph
        self.build_word_graph(window_size)

        # Compute the word weights using the unweighted PageRank
        # formulae
        weights = nx.pagerank_scipy(self.graph,
                                    alpha=0.85,
                                    tol=0.0001,
                                    weight=None)

        # Generate the phrases from the T-percent top weighted words
        if top_t_percent is not None:

            # Computing the number of top keywords
            number_of_nodes = self.graph.number_of_nodes()
            to_keep = int(number_of_nodes * top_t_percent)

            # Sorting the nodes by decreasing weights
            sorted_weights = sorted(weights, key=weights.get, reverse=True)

            # Creating keyphrases from the T top words
            self.select_candidates_with_longest_keyword_sequences(
                keywords=sorted_weights[:int(to_keep)])

        self.weight_candidates_with_words_weights(weights, normalize_weights)

    def weight_candidates_with_words_weights(self,
                                             weights,
                                             normalize_weights,
                                             use_position_adjustment=True):
        """
        Weights candidates using the sum of their word weights.

        Parameters
        ----------
        weights: dict
            Word weights

        normalize_weights: bool
            Whether normalize keyphrase weight by their length.

        use_position_adjustment: bool
            Whether use candidate position to adjust weights,
            defaults to `True`.
        """

        for candidate in self.candidates.values():
            candidate.weight = sum([weights.get(word, 0.0)
                                    for word in candidate.normalized_words])

            if normalize_weights:
                candidate.weight /= len(candidate.normalized_words)

            if use_position_adjustment:
                candidate.weight += candidate.offsets[0] * 1e-8



from itertools import combinations

import networkx as nx
import numpy as np
from scipy.cluster.hierarchy import linkage, fcluster
from scipy.spatial.distance import pdist

from perke.base.extractor import Extractor
from perke.base.types import (TopicHeuristic,
                              HierarchicalClusteringMetric,
                              HierarchicalClusteringLinkageMethod)


class TopicRank(Extractor):
    """
    TopicRank keyphrase extractor.

    This model relies on a topical representation of the text. Candidate
    keyphrases are clustered into topics and used as nodes in a complete
    graph. A graph-based ranking model is applied to assign a
    significance weight to each topic. Keyphrases are then generated by
    selecting a candidate from each of the top ranked topics.

    Implementation of the SingleRank model described in:

    Adrien Bougouin, Florian Boudin and Béatrice Daille.
    TopicRank: Graph-Based Topic Ranking for Keyphrase Extraction.
    In proceedings of IJCNLP, pages 543-551, 2013.

    Examples
    --------
    from perke.unsupervised.graph_based import TopicRank
    from perke.base.types import (WordNormalizationMethod,
                                  HierarchicalClusteringMetric,
                                  HierarchicalClusteringLinkageMethod)

    # Define the set of valid part of speech tags to occur in the model.
    valid_pos_tags = {'N', 'Ne', 'AJ', 'AJe'}

    # 1. Create a TopicRank extractor.
    extractor = TopicRank(valid_pos_tags=valid_pos_tags)

    # 2. Load the text.
    extractor.load_text(
        input='text or path/to/input_file',
        word_normalization_method=WordNormalizationMethod.stemming)

    # 3. Select the longest sequences of nouns and adjectives, that do
    #    not contain punctuation marks or stopwords as candidates.
    extractor.select_candidates()

    # 4. Build topics by grouping candidates with HAC (average linkage,
    #    jaccard distance, threshold of 1/4 of shared normalized words).
    #    Weight the topics using random walk, and select the first
    #    occurring candidate from each topic.
    extractor.weight_candidates(
        threshold=0.74,
        metric=HierarchicalClusteringMetric.jaccard,
        linkage_method=HierarchicalClusteringLinkageMethod.average)

    # 5. Get the 10 highest weighted candidates as keyphrases
    keyphrases = extractor.get_n_best(n=10)

    Attributes
    ----------
    graph: nx.Graph
        The topic graph

    topics: list
        List of topics
    """

    def __init__(self, valid_pos_tags=None):
        """
        Initializes TopicRank.

        Parameters
        ----------
        valid_pos_tags: set
            Set of valid part of speech tags, defaults to nouns and
            adjectives. I.e. `{'N', 'Ne', 'AJ', 'AJe'}`.
        """

        super().__init__(valid_pos_tags)
        self.graph = nx.Graph()
        self.topics = []

    def select_candidates(self):
        """
        Selects longest sequences of nouns and adjectives as keyphrase
        candidates.
        """

        # Select sequence of adjectives and nouns
        self.select_candidates_with_longest_pos_sequences(
            valid_pos_tags=self.valid_pos_tags)

        # Filter candidates containing stopwords or punctuation marks
        self.filter_candidates(stopwords=self.stopwords)

    def vectorize_candidates(self):
        """
        Vectorize the keyphrase candidates.

        Returns
        -------
        candidates: list
            The list of candidates (canonical forms).

        candidate_matrix: np.ndarray
            Vectorized representation of the candidates.
        """

        # Build the vocabulary, i.e. setting the vector dimensions
        vocabulary = set()
        for candidate in self.candidates.values():
            for word in candidate.normalized_words:
                vocabulary.add(word)
        vocabulary = list(vocabulary)

        # Vectorize the candidates and sort for random issues
        candidates = list(self.candidates)
        candidates.sort()

        candidate_matrix = np.zeros((len(candidates), len(vocabulary)))
        for i, c in enumerate(candidates):
            for word in self.candidates[c].normalized_words:
                candidate_matrix[i, vocabulary.index(word)] += 1

        return candidates, candidate_matrix

    def cluster_topics(
            self,
            threshold=0.74,
            metric=HierarchicalClusteringMetric.jaccard,
            linkage_method=HierarchicalClusteringLinkageMethod.average):
        """
        Clusters candidates into topics.

        Parameters
        ----------
        threshold: float
            The minimum similarity for clustering, defaults to `0.74`,
            i.e. more than 1/4 of normalized word overlap similarity.

        metric: str
            The hierarchical clustering metric, defaults to `'jaccard'`
            See `HierarchicalClusteringMetric` for available methods.

        linkage_method: str
            The hierarchical clustering linkage method, defaults to
            `'average'`. See HierarchicalClusteringLinkageMethod
            for available methods.
        """

        # Handle content with only one candidate
        if len(self.candidates) == 1:
            self.topics.append(list(self.candidates))
            return

        # Vectorize the candidates
        candidates, candidate_matrix = self.vectorize_candidates()

        # Compute the distance matrix
        distance_matrix = pdist(candidate_matrix, metric)
        distance_matrix = np.nan_to_num(distance_matrix)

        # Compute the clusters
        clusters = linkage(distance_matrix, method=linkage_method)

        # Form flat clusters
        flat_clusters = fcluster(clusters, t=threshold, criterion='distance')

        # For each topic identifier
        for cluster_id in range(1, max(flat_clusters) + 1):
            self.topics.append([candidates[j] for j in range(len(flat_clusters))
                                if flat_clusters[j] == cluster_id])

    def build_topic_graph(self):
        """
        Build topic graph.
        """

        # Adding the nodes to the graph
        self.graph.add_nodes_from(range(len(self.topics)))

        # Loop through the topics to connect the nodes
        for i, j in combinations(range(len(self.topics)), 2):
            self.graph.add_edge(i, j, weight=0.0)
            for c_i in self.topics[i]:
                candidate_i = self.candidates[c_i]
                for c_j in self.topics[j]:
                    candidate_j = self.candidates[c_j]
                    for p_i in candidate_i.offsets:
                        for p_j in candidate_j.offsets:
                            gap = abs(p_i - p_j)
                            if p_i < p_j:
                                gap -= len(candidate_i.normalized_words) - 1
                            elif p_j < p_i:
                                gap -= len(candidate_j.normalized_words) - 1

                            self.graph[i][j]['weight'] += 1.0 / gap

    def weight_candidates(
            self,
            threshold=0.74,
            metric=HierarchicalClusteringMetric.jaccard,
            linkage_method=HierarchicalClusteringLinkageMethod.average,
            topic_heuristic=TopicHeuristic.first_occurring):
        """
        Candidate ranking using random walk.

        Parameters
        ----------
        threshold: float
            The minimum similarity for clustering, defaults to 0.74.

        metric: str
            The hierarchical clustering metric, defaults to `'jaccard'`
            See `HierarchicalClusteringMetric` for available methods.

        linkage_method: str
            The hierarchical clustering linkage method, defaults to
            `'average'`. See HierarchicalClusteringLinkageMethod for
            available methods.

        topic_heuristic: str
            The heuristic for selecting the best candidate for each
            topic, defaults to first occurring candidate. See
            `TopicHeuristic` for available heuristics.
        """

        # Cluster the candidates
        self.cluster_topics(threshold, metric, linkage_method)

        # Build the topic graph
        self.build_topic_graph()

        # Compute the word weights using random walk
        weights = nx.pagerank_scipy(self.graph, alpha=0.85, weight='weight')

        # Loop through the topics
        for i, topic in enumerate(self.topics):

            # Get the offsets of the topic candidates
            offsets = [self.candidates[c].offsets[0] for c in topic]

            # Get first candidate from topic
            if topic_heuristic == TopicHeuristic.frequent:

                # Get frequencies for each candidate within the topic
                frequencies = [len(self.candidates[c].all_words)
                               for c in topic]

                # Get the indices of the most frequent candidates
                max_frequency = max(frequencies)
                indices = [j for j, f in enumerate(frequencies)
                           if f == max_frequency]

                # Offsets of the indexes
                indices_offsets = [offsets[j] for j in indices]
                most_frequent = indices_offsets.index(min(indices_offsets))
                self.candidates[topic[most_frequent]].weight = weights[i]

            else:
                first = offsets.index(min(offsets))
                self.candidates[topic[first]].weight = weights[i]



from collections import defaultdict
import logging

import nltk
import hazm

from perke.base.data_structures import Candidate
from perke.base.readers import RawTextReader
from perke.base.types import WordNormalizationMethod
from perke.base.functions import is_alphanumeric
from perke.base.string import punctuation_marks


class Extractor:
    """
    Base extractor, provides base functions for all extractors.

    Attributes
    ----------
    word_normalization_method: str
        Word normalization method, See `WordNormalizationMethod` for
        methods.

    sentences: list
        List of sentence objects of the text

    candidates: defaultdict
        Dict of canonical forms of candidates to candidates, canonical
        form of a candidate is a string joined from normalized words of
        the candidate.

    stopwords: set
        Set of stopwords

    valid_pos_tags: set
        Set of valid part of speech tags.
    """

    def __init__(self, valid_pos_tags=None):
        """
        Initializes the extractor.

        Parameters
        ----------
        valid_pos_tags: set
            Set of valid part of speech tags, defaults to nouns and
            adjectives. I.e. `{'N', 'Ne', 'AJ', 'AJe'}`.
        """
        self.word_normalization_method = None
        self.sentences = []
        self.candidates = defaultdict(Candidate)
        self.stopwords = set(hazm.stopwords_list()) | set(punctuation_marks)
        if valid_pos_tags is None:
            self.valid_pos_tags = {'N', 'Ne', 'AJ', 'AJe'}
        else:
            self.valid_pos_tags = valid_pos_tags

    def load_text(self,
                  input,
                  word_normalization_method=WordNormalizationMethod.stemming):
        """
        Loads the text of a document or string.

        Parameters
        ----------
        input: str
            Input, this can be either raw text or filepath.

        word_normalization_method: str
            Word normalization method, defaults to `'stemming'`. See
            `WordNormalizationMethod` for available methods.
        """

        # Initialize reader
        reader = RawTextReader(input, word_normalization_method)

        # Load sentences
        self.sentences = reader.read()

        self.word_normalization_method = word_normalization_method

    def is_redundant(self, candidate, selected_candidates, minimum_length=1):
        """
        Test if a candidate is redundant with respect to a list of
        already selected candidates. A candidate is considered redundant
        if it is included in another candidate that is weighted higher
        in the list.

        Parameters
        ----------
        candidate: str
            the canonical form of the candidate

        selected_candidates: list
            the list of already selected candidates canonical forms

        minimum_length: int
            minimum length of the candidate to be considered,
            defaults to `1`.

        Returns
        -------
        result: bool
            The result
        """

        # Only consider candidate with length greater than minimum
        # length
        if self.candidates[candidate].length < minimum_length:
            return False

        # Loop through the already selected candidates
        for sc in selected_candidates:
            if candidate in sc:
                return True

        return False

    def get_n_best(self,
                   n=10,
                   remove_redundants=False,
                   normalized=False):
        """
        Returns the n best candidates.

        Parameters
        ----------
        n: int
            The number of candidates, defaults to `10`.

        remove_redundants: bool
            Whether redundant keyphrases are filtered out from the
            n-best list, defaults to `False`.

        normalized: bool
            Whether to get normalized words instead of words of first
            occurring form of candidate, defaults to `False`.

        Returns
        -------
        n_best: list
            List of `(candidate, weight)` tuples, `candidate` can be
            either canonical form or first occurrence joined words.
        """

        # Sort candidates by descending weight
        bests = sorted(self.candidates,
                       key=lambda c: self.candidates[c].weight,
                       reverse=True)

        # Remove redundant candidates
        if remove_redundants:

            # Initialize a new container for non redundant candidates
            non_redundant_bests = []

            # Loop through the best candidates
            for c in bests:

                # Test whether candidate is redundant
                if self.is_redundant(candidate=c,
                                     selected_candidates=non_redundant_bests):
                    continue

                # Add the candidate otherwise
                non_redundant_bests.append(c)

                # Break computation if the n-best are found
                if len(non_redundant_bests) >= n:
                    break

            # Copy non redundant candidates in best container
            bests = non_redundant_bests

        # Get the list of best candidate
        n_best = []
        for c in bests[:min(n, len(bests))]:
            candidate = self.candidates[c]
            if normalized:
                # (canonical form, weight) tuples
                n_best.append((c, candidate.weight))
            else:
                # (first occurrence joined words, weight) tuples
                n_best.append((' '.join(candidate.all_words[0]),
                               candidate.weight))

        if len(n_best) < n:
            logging.warning('Not enough candidates to choose from '
                            '({} requested, {} given).'.format(n, len(n_best)))

        return n_best

    def add_candidate_occurrence(self,
                                 words,
                                 offset,
                                 pos_tags,
                                 normalized_words):
        """
        Adds a new candidate occurrence.

        Parameters
        ----------
        words: list
            List of words of the occurrence

        normalized_words: list
            List of normalized of words of the occurrence

        pos_tags: list
            List of part of speech tags assigned to words of the
            occurrence

        offset: int
            The offset of the occurrence
        """

        # Build the canonical form of the candidate
        canonical_form = ' '.join(normalized_words)

        # Create candidate if not exist and add occurrence
        self.candidates[canonical_form].add_occurrence(words,
                                                       offset,
                                                       pos_tags,
                                                       normalized_words)

    def select_candidates_with_longest_pos_sequences(self, valid_pos_tags):
        """
        Selects candidates with longest sequences of given part of
        speech tags.

        Parameters
        ----------
        valid_pos_tags: set
            Set of valid part of speech tags
        """
        self.select_candidates_with_longest_sequences(
            key=lambda sentence: sentence.pos_tags,
            valid_values=valid_pos_tags)

    def select_candidates_with_longest_keyword_sequences(self, keywords):
        """
        Selects candidates with longest sequences of given keywords.

        Parameters
        ----------
        keywords: set
            Set of given keywords
        """
        self.select_candidates_with_longest_sequences(
            key=lambda sentence: sentence.normalized_words,
            valid_values=keywords)

    def select_candidates_with_longest_sequences(self, key, valid_values):
        """
        Selects candidates with longest sequences of given values, based
        on `key`.

        Parameters
        ----------
        key: function
            Function that given a sentence and returns a list

        valid_values: set
            The valid values
        """

        # Loop through the sentences
        offset_shift = 0
        for i, sentence in enumerate(self.sentences):
            sequence_offsets = []

            # Loop through the key result
            for j, value in enumerate(key(sentence)):

                # Add candidate offset in sequence and continue if not
                # last word
                if value in valid_values:
                    sequence_offsets.append(j)
                    if j < (sentence.length - 1):
                        continue

                # Add sequence as candidate if it is not empty
                if len(sequence_offsets) > 0:
                    first = sequence_offsets[0]
                    last = sequence_offsets[-1]

                    # Add the ngram as a new candidate occurrence
                    self.add_candidate_occurrence(
                        words=sentence.words[first: last + 1],
                        offset=offset_shift + first,
                        pos_tags=sentence.pos_tags[first: last + 1],
                        normalized_words=sentence.normalized_words[first: last + 1])

                # Flush sequence offsets
                sequence_offsets = []

            offset_shift += sentence.length

    def select_candidates_with_grammar(self, grammar=None):
        """
        Selects candidates using nltk RegexpParser with a grammar
        defining noun phrases (NP).

        Parameters
        ----------
        grammar: str
            grammar defining part of speech patterns of noun phrases,
            defaults to:
            `'NP:
                  <P>{<N>}<V>
              NP:
                  {<DETe?|Ne?|NUMe?|AJe|PRO|CL|RESe?><DETe?|Ne?|NUMe?
                  |AJe?|PRO|CL|RESe?>*}
                  <N>}{<.*e?>'`
        """

        # Initialize default grammar if none provided
        if grammar is None:
            grammar = r"""
                NP:
                    <P>{<N>}<V>
                NP:
                    {<DETe?|Ne?|NUMe?|AJe|PRO|CL|RESe?><DETe?|Ne?|NUMe?|AJe?
                    |PRO|CL|RESe?>*}
                    <N>}{<.*e?>
            """

        # Initialize parser
        parser = nltk.RegexpParser(grammar)

        # Loop through the sentences
        offset_shift = 0
        for i, sentence in enumerate(self.sentences):
            # Convert sentence as list of (offset, pos) tuples
            tuples = [(str(j), sentence.pos_tags[j])
                      for j in range(sentence.length)]

            # Parse sentence
            tree = parser.parse(tuples)

            # Find candidates
            for subtree in tree.subtrees():
                if subtree.label() == 'NP':
                    leaves = subtree.leaves()

                    # Get the first and last offset of the current
                    # candidate
                    first = int(leaves[0][0])
                    last = int(leaves[-1][0])

                    # Add the noun phrase to the candidate container
                    self.add_candidate_occurrence(
                        words=sentence.words[first:last + 1],
                        offset=offset_shift + first,
                        pos_tags=sentence.pos_tags[first:last + 1],
                        normalized_words=sentence.normalized_words[first:last + 1])

            # Compute offset shift
            offset_shift += sentence.length

    def filter_candidates(self,
                          stopwords=None,
                          minimum_characters=3,
                          minimum_word_characters=2,
                          valid_punctuation_marks='-',
                          maximum_length=5,
                          alphanumeric_only=True,
                          invalid_pos_tags=None):
        """
        Filters the candidates with given conditions.

        Parameters
        ----------
        stopwords: set
            Set of stopwords, defaults to an empty set.

        minimum_characters: int
            Minimum number of characters for a candidate, defaults to
            `3`.

        minimum_word_characters: int
            Minimum number of characters for a word to be considered as
            a valid word, defaults to `2`.

        valid_punctuation_marks: str
            Punctuation marks that are valid for a candidate, defaults
            to `'-'`.

        maximum_length: int
            Maximum length in words of the candidate, defaults to `5`.

        alphanumeric_only: bool
            Filters candidates containing non alpha-numeric characters,
            defaults to `True`.

        invalid_pos_tags: set
            Set of unwanted part of speech tags in candidates, defaults
            to an empty set.
        """

        if stopwords is None:
            stopwords = set()

        if invalid_pos_tags is None:
            invalid_pos_tags = set()

        # Loop through the candidates
        for c in list(self.candidates):

            # Get the candidate
            candidate = self.candidates[c]

            # Get the words from the first occurrence
            words = candidate.all_words[0]

            # Discard if words are in the stoplist
            if set(words).intersection(stopwords):
                del self.candidates[c]

            # Discard if any of pos tags are in the invalid pos tags
            elif set(candidate.all_pos_tags[0]).intersection(invalid_pos_tags):
                del self.candidates[c]

            # Discard if containing words composed of only punctuation
            elif any([set(word).issubset(set(punctuation_marks))
                      for word in words]):
                del self.candidates[c]

            # Discard short candidates
            elif len(''.join(words)) < minimum_characters:
                del self.candidates[c]

            # Discard candidates containing short words
            elif min([len(word) for word in words]) < minimum_word_characters:
                del self.candidates[c]

            # Discard candidates with long length
            elif len(candidate.normalized_words) > maximum_length:
                del self.candidates[c]

            # Discard if not containing only alpha-numeric characters
            elif alphanumeric_only and not all([is_alphanumeric(word,
                                                                valid_punctuation_marks)
                                                for word in words]):
                del self.candidates[c]



import networkx as nx

from perke.unsupervised.graph_based.text_rank import TextRank


class SingleRank(TextRank):
    """
    SingleRank keyphrase extractor

    This model is an extension of the TextRank model that uses the
    number of co-occurrences to weight edges in the graph.

    Implementation of the SingleRank model described in:

    Xiaojun Wan and Jianguo Xiao.
    Single Document Keyphrase Extraction Using Neighborhood Knowledge.
    In proceedings of the NCAI, pages 855–860, 2008.

    Examples
    --------
    from perke.unsupervised.graph_based import SingleRank

    # Define the set of valid part of speech tags to occur in the model.
    valid_pos_tags = {'N', 'Ne', 'AJ', 'AJe'}

    # 1. Create a SingleRank extractor.
    extractor = SingleRank(valid_pos_tags=valid_pos_tags)

    # 2. Load the text.
    extractor.load_text(input='text or path/to/input_file',
                        word_normalization_method=None)

    # 3. Select the longest sequences of nouns and adjectives as
    #    candidates.
    extractor.select_candidates()

    # 4. Weight the candidates using the sum of their words weights that
    #    are computed using random walk. In the graph, nodes are certain
    #    parts of speech (nouns and adjectives) that are connected if
    #    they co-occur in a window of 10 words.
    extractor.weight_candidates(window=10)

    # 5. Get the 10 highest weighted candidates as keyphrases
    keyphrases = extractor.get_n_best(n=10)
    """

    def __init__(self, valid_pos_tags=None):
        """
        initializes SingleRank.

        Parameters
        ----------
        valid_pos_tags: set
            Set of valid part of speech tags, defaults to nouns and
            adjectives. I.e. `{'N', 'Ne', 'AJ', 'AJe'}`.
        """
        super().__init__(valid_pos_tags)
        self.graph_edges_are_weighted = True

    def build_word_graph(self, window_size=10):
        """
        Builds a graph representation of the text just like TextRank
        except that edges in SingleRank graph are weighted and the
        number of times two words co-occur in a window is encoded as
        edge weights.

        Parameters
        ----------
        window_size: int
            The size of window for connecting two words in the graph,
            defaults to `10`.
        """
        super().build_word_graph(window_size)

    def weight_candidates(self,
                          window_size=10,
                          normalize_weights=False,
                          **kwargs):
        """
        Weights candidates using the weighted variant of the TextRank
        formulae. Candidates are weighted by the sum of the weights of
        their words.

        Parameters
        ----------
        window_size: int
            The size of window for connecting two words in the graph,
            defaults to `10`.

        normalize_weights: bool
            Whether normalize keyphrase weight by their length, defaults
            to `False`.
        """

        # Build the word graph
        self.build_word_graph(window_size)

        # Compute the word weights using random walk
        weights = nx.pagerank_scipy(self.graph,
                                    alpha=0.85,
                                    tol=0.0001,
                                    weight='weight')

        self.weight_candidates_with_words_weights(weights, normalize_weights)



from os.path import (join,
                     dirname,
                     isfile)

import hazm

from perke.base.types import WordNormalizationMethod
from perke.base.data_structures import Sentence


class Reader:
    """
    Base Reader

    Attributes
    ----------
    word_normalization_method: str
        Word normalization method

    normalizer: hazm.Normalizer
        The hazm normalizer instance

    stemmer: hazm.Stemmer
        The hazm stemmer instance

    lemmatizer: hazm.Lemmatizer
        The hazm lemmatizer instance

    pos_tagger: hazm.POSTagger
        The hazm pos tagger instance
    """

    def __init__(self, word_normalization_method):
        """
        Initializes the reader.

        Parameters
        ----------
        word_normalization_method: str
            Word normalization method, see `WordNormalizationMethod` for
            available methods.
        """

        self.word_normalization_method = word_normalization_method
        self.normalizer = hazm.Normalizer()
        self.stemmer = hazm.Stemmer()
        self.lemmatizer = hazm.Lemmatizer()
        model_filepath = join(dirname(dirname(__file__)),
                              'resources',
                              'postagger.model')
        self.pos_tagger = hazm.POSTagger(model=model_filepath)


class RawTextReader(Reader):
    """
    Reader for raw text

    Attributes
    ----------
    text: str
        Raw text to read sentences from
    """

    def __init__(self, input, word_normalization_method):
        """
        Initializes the reader.

        Parameters
        ----------
        input: str
            Input, this can be either raw text or filepath.

        word_normalization_method: str
            Word normalization method, see NormalizationMethod for
            available methods.
        """

        super().__init__(word_normalization_method)

        # If input is a filepath
        if isfile(input):
            with open(input) as file:
                self.text = file.read()

        # If input is raw text
        else:
            self.text = input

    def read(self):
        """
        Reads the input and uses hazm to preprocess.

        Returns
        -------
        sentences: list
            List of sentences
        """
        word_normalization_method = self.word_normalization_method
        normalized_text = self.normalizer.normalize(self.text)
        sentences = []
        for sentence in hazm.sent_tokenize(normalized_text):
            words = hazm.word_tokenize(sentence)
            pos_tags = [tag for _, tag in self.pos_tagger.tag(words)]

            if word_normalization_method == WordNormalizationMethod.stemming:
                normalized_words = [self.stemmer.stem(word) for word in words]

            elif word_normalization_method == WordNormalizationMethod.lemmatization:
                normalized_words = [self.lemmatizer.lemmatize(word)
                                    for word in words]

            # No normalization
            else:
                normalized_words = words

            sentence = Sentence(words, pos_tags, normalized_words)
            sentences.append(sentence)

        return sentences



from os.path import join
from setuptools import (setup,
                        find_packages)

packages = find_packages(exclude=['tests*'])

with open('requirements.txt') as f:
    requirements = f.read().split()

with open('README.md') as f:
    long_description = f.read()

filename = join('perke', 'version.py')
with open(filename) as f:
    exec(compile(f.read(), filename, 'exec'))


setup(name='perke',
      version=__version__,
      description='A keyphrase extractor for Persian',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/alirezah320/perke',
      author='Alireza Hosseini',
      author_email='alirezah320@yahoo.com',
      packages=packages,
      license='MIT',
      install_requires=requirements,
      python_requires='>=3.6',
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
      ])



import math
from itertools import combinations

import networkx as nx

from perke.unsupervised.graph_based.topic_rank import TopicRank
from perke.base.types import (HierarchicalClusteringMetric,
                              HierarchicalClusteringLinkageMethod)


class MultipartiteRank(TopicRank):
    """
    MultipartiteRank keyphrase extractor

    This model encodes topical information within a multipartite graph
    structure. The model represents keyphrase candidates and topics in
    a single graph and exploits their mutually reinforcing relationship
    to improve candidate ranking.

    Implementation of the PositionRank described in:

    Florian Boudin.
    Unsupervised Keyphrase Extraction with Multipartite Graphs.
    In proceedings of NAACL, pages 667-672, 2018.

    Examples
    --------
    from perke.unsupervised.graph_based import MultipartiteRank
    from perke.base.types import (WordNormalizationMethod,
                                  HierarchicalClusteringMetric,
                                  HierarchicalClusteringLinkageMethod)

    # Define the set of valid part of speech tags to occur in the model.
    valid_pos_tags = {'N', 'Ne', 'AJ', 'AJe'}

    # 1. Create a MultipartiteRank extractor.
    extractor = MultipartiteRank(valid_pos_tags=valid_pos_tags)

    # 2. Load the text.
    extractor.load_text(
        input='text or path/to/input_file',
        word_normalization_method=WordNormalizationMethod.stemming)

    # 3. Select the longest sequences of nouns and adjectives, that do
    #    not contain punctuation marks or stopwords as candidates.
    extractor.select_candidates()

    # 4. Build the Multipartite graph and weight candidates using
    #    random walk, alpha controls the weight adjustment mechanism,
    #    see TopicRank for metric, linkage method and threshold
    #    parameters.
    extractor.weight_candidates(
        threshold=0.74,
        metric=HierarchicalClusteringMetric.jaccard,
        linkage_method=HierarchicalClusteringLinkageMethod.average,
        alpha=1.1)

    # 5. Get the 10 highest weighted candidates as keyphrases
    keyphrases = extractor.get_n_best(n=10)

    Attributes
    ----------
    topic_ids: dict
        Dict of canonical forms of candidates to to topic identifiers

    graph: nx.DiGraph
        The candidate graph
    """

    def __init__(self, valid_pos_tags=None):
        """
        Initializes  MultipartiteRank.

        Parameters
        ----------
        valid_pos_tags: set
            Set of valid part of speech tags, defaults to nouns and
            adjectives. I.e. `{'N', 'Ne', 'AJ', 'AJe'}`.
        """

        super().__init__(valid_pos_tags)
        self.topic_ids = {}
        self.graph = nx.DiGraph()

    def cluster_topics(
            self,
            threshold=0.74,
            metric=HierarchicalClusteringMetric.jaccard,
            linkage_method=HierarchicalClusteringLinkageMethod.average):
        """
        Clustering candidates into topics.

        Parameters
        ----------
        threshold: float
            The minimum similarity for clustering, defaults to `0.74`,
            i.e. more than 1/4 of normalized word overlap similarity.

        metric: str
            The hierarchical clustering metric, defaults to `'jaccard'`
            See `HierarchicalClusteringMetric` for available methods.

        linkage_method: str
            The hierarchical clustering linkage method, defaults to
            `'average'`. See HierarchicalClusteringLinkageMethod
            for available methods.
        """
        super().cluster_topics(threshold, metric, linkage_method)

        # Handle text with only one candidate
        if len(self.candidates) == 1:
            self.topic_ids[list(self.candidates)[0]] = 0
            return

        # Assign cluster identifiers to candidates
        for topic_id, topic in enumerate(self.topics):
            for c in topic:
                self.topic_ids[c] = topic_id

    def build_topic_graph(self):
        """
        Build the Multipartite graph.
        """

        # Adding the nodes to the graph
        self.graph.add_nodes_from(self.candidates.keys())

        # Pre-compute edge weights
        for node_i, node_j in combinations(self.candidates.keys(), 2):

            # Discard intra-topic edges
            if self.topic_ids[node_i] == self.topic_ids[node_j]:
                continue

            weights = []
            candidate_i = self.candidates[node_i]
            candidate_j = self.candidates[node_j]
            for p_i in candidate_i.offsets:
                for p_j in candidate_j.offsets:

                    # Compute gap
                    gap = abs(p_i - p_j)

                    # Alter gap according to candidate length
                    if p_i < p_j:
                        gap -= len(candidate_i.normalized_words) - 1
                    if p_j < p_i:
                        gap -= len(candidate_j.normalized_words) - 1

                    weights.append(1.0 / gap)

            # Add weighted edges
            if weights:
                # node_i -> node_j
                self.graph.add_edge(node_i, node_j, weight=sum(weights))
                # node_j -> node_i
                self.graph.add_edge(node_j, node_i, weight=sum(weights))

    def weight_adjustment(self, alpha=1.1):
        """
        Adjust edge weights for boosting some candidates.

        Parameters
        ----------
        alpha: float
            Hyper-parameter that controls the strength of the weight
            adjustment, defaults to `1.1`.
        """

        weighted_edges = {}

        # Topical boosting
        for topic in self.topics:

            # Skip one candidate topics
            if len(topic) == 1:
                continue

            # Get the offsets
            offsets = [self.candidates[c].offsets[0] for c in topic]

            # Get the first occurring candidate
            first = topic[offsets.index(min(offsets))]

            # Find the nodes to which it connects
            for start, end in self.graph.edges(first):

                boosters = []
                for c in topic:
                    if c != first and self.graph.has_edge(c, end):
                        boosters.append(self.graph[c][end]['weight'])

                if boosters:
                    weighted_edges[(start, end)] = sum(boosters)

        # Update edge weights
        for nodes, boosters in weighted_edges.items():
            node_i, node_j = nodes
            position_i = 1.0 / (1 + self.candidates[node_i].offsets[0])
            position_i = math.exp(position_i)
            self.graph[node_j][node_i]['weight'] += boosters * alpha * position_i

    def weight_candidates(
            self,
            threshold=0.74,
            metric=HierarchicalClusteringMetric.jaccard,
            linkage_method=HierarchicalClusteringLinkageMethod.average,
            alpha=1.1):
        """
        Candidate weight calculation using random walk.

        Parameters
        ----------
        threshold: float
            The minimum similarity for clustering, defaults to 0.74.

        metric: str
            The hierarchical clustering metric, defaults to `'jaccard'`
            See `HierarchicalClusteringMetric` for available methods.

        linkage_method: str
            The hierarchical clustering linkage method, defaults to
            `'average'`. See HierarchicalClusteringLinkageMethod for
            available methods.

        alpha: float
            Hyper-parameter that controls the strength of the
            weight adjustment, defaults to `1.1`
        """

        # Cluster the candidates
        self.cluster_topics(threshold=threshold, linkage_method=linkage_method)

        # Build the topic graph
        self.build_topic_graph()

        if alpha > 0.0:
            self.weight_adjustment(alpha)

        # Compute the word weights using random walk
        weights = nx.pagerank_scipy(self.graph)
        for c in weights:
            self.candidates[c].weight = weights[c]